namespace BetDriver.StrategyAutotrader
{
	using System;
	using System.Collections.Generic;
	using System.ComponentModel;
	using System.IO;
	using System.Linq;

	using BetDriver.Api;
	using BetDriver.Api.Common;
	using BetDriver.Api.Localization;
	using BetDriver.Api.Rules;
	using BetDriver.Api.Strategies;
	using BetDriver.Api.TO;
	using BetDriver.Api.Xaml;
	using BetDriver.Api.Properties;

	using Xceed.Wpf.Toolkit.PropertyGrid.Attributes;

	using MarketStatus = BetDriver.Api.MarketStatus;

	[LocalizedDisplayName(typeof(Resources),"RSC_AvtoTreyder")]
	public class StrategyAutotrader : SearchAutoStrategy
    {
		public StrategyAutotrader()
			: base(true)
		{
			RuleSet = new RuleSet { new AutotraderRule() };
			Description = Resources.RSC_Strategiyaperestavlyaetvse;
		}

		/// <summary>
		/// Проверить возможно ли удалить рынок.
		/// </summary>
		/// <param name="market">Рынок.</param>
		/// <returns>True, если рынок может быть удален, иначе false.</returns>
		protected override bool OnCanRemove(Market market)
		{
			return RuleSet.OfType<AutotraderRule>().First().TryRemoveMarket(market, false);
		}

		/// <summary>
		/// Удалить рынок из обработки.
		/// </summary>
		/// <param name="market">Рынок.</param>
		protected override void OnRemoveMarket(Market market)
		{
			RuleSet.OfType<AutotraderRule>().First().TryRemoveMarket(market, true);
		}
    }

	public enum StopActivationType
	{
		/// <summary>
		/// Текущий коэффициент 
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Tekushchiykoeffitsient")]
		Koeff,

		/// <summary>
		/// Прошла сделка
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Proshlasdelka")]
		Trade
	}

	/// <summary>
	/// Автотрейдер.
	/// </summary>
	[LocalizedDisplayName(typeof(Resources),"RSC_AvtoTreyder")]
	public class AutotraderRule : Rule
	{
		private sealed class BetInfo
		{
			public Bet Bet { get; set; }

			public Bet SourceBet { get; private set; }

			public bool NeedStop { get; set; }

			public bool NeedPlace { get; set; }

			public int BetsCount { get; set; }

			public BetInfo(Bet bet)
			{
				if (bet == null)
					throw new ArgumentNullException("bet");

				SourceBet = bet;
				NeedPlace = true;
			}
		}

		private readonly HashSet<Bet> _processedBets = new HashSet<Bet>();
		private readonly Dictionary<Bet, BetInfo> _bets = new Dictionary<Bet, BetInfo>();

		private readonly HashSet<Market> _marketsToRemove = new HashSet<Market>();
		private readonly object _sync = new object();

		private DateTime _lastCleanOldMarkets;

		#region Параметры

		private bool _closeMarkets;
		private TimeSpan _closeRange;
		private TimeSpan _closePeriod;
		private bool _cancelAllBetsInPlay;
		private decimal _ignoreIfWinOnClose;
		private decimal _ignoreBetAmountOnClose;
		private bool _lay;
		private Margin _layOffset;
		private bool _back;
		private Margin _backOffset;
		private bool _attachBets;
		private decimal _profitOfClosedMarkets;
		private int _stopValue;
		private int _stopSafeSpread;
		private StopActivationType _stopActivationType;
		private decimal _stopMinimumAmountOnPrice;
		private int _maxBetsCount;

		/// <summary>
		/// При принятии ставки Против делать обратную ставку За.
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_PerestavlyatProtiv")]
		[LocalizedDescription(typeof(Resources),"RSC_PriprinyatiistavkiProtivde")]
		[LocalizedCategory(typeof(Resources),"RSC_2Obrabotkarblnkov")]
		[Parameter]
		[PropertyOrder(40)]
		public bool Lay
		{
			get { return _lay; }
			set
			{
				_lay = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Сдвиг для Против.
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_SdvigstavokZa")]
		[LocalizedDescription(typeof(Resources),"RSC_SdvigstavokZa")]
		[LocalizedCategory(typeof(Resources),"RSC_2Obrabotkarblnkov")]
		[Parameter]
		[PropertyOrder(50)]
		public Margin LayOffset
		{
			get { return _layOffset; }
			set
			{
				if (_layOffset != null)
					_layOffset.PropertyChanged -= OnPropertyChanged;

				_layOffset = value;

				if (_layOffset != null)
					_layOffset.PropertyChanged += OnPropertyChanged;

				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// При принятии ставки За делать обратную ставку Против.
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_PerestavlyatZa")]
		[LocalizedDescription(typeof(Resources),"RSC_PriprinyatiistavkiZadelato")]
		[LocalizedCategory(typeof(Resources),"RSC_2Obrabotkarblnkov")]
		[Parameter]
		[PropertyOrder(60)]
		public bool Back
		{
			get { return _back; }
			set
			{
				_back = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Сдвиг для За.
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_SdvigstavokProtiv")]
		[LocalizedDescription(typeof(Resources),"RSC_SdvigstavokProtiv")]
		[LocalizedCategory(typeof(Resources),"RSC_2Obrabotkarblnkov")]
		[Parameter]
		[PropertyOrder(70)]
		public Margin BackOffset
		{
			get { return _backOffset; }
			set
			{
				if (_backOffset != null)
					_backOffset.PropertyChanged -= OnPropertyChanged;

				_backOffset = value;

				if (_backOffset != null)
					_backOffset.PropertyChanged += OnPropertyChanged;

				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Подключать ставки, сделанные извне программы.
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Podklyuchatstavkiizvne")]
		[LocalizedDescription(typeof(Resources),"RSC_Podklyuchatstavkisdelannbl")]
		[LocalizedCategory(typeof(Resources),"RSC_2Obrabotkarblnkov")]
		[Parameter]
		[PropertyOrder(80)]
		public bool AttachBets
		{
			get { return _attachBets; }
			set
			{
				_attachBets = value;
				RaiseSettingsChanged();
			}
		}

		[LocalizedDisplayName(typeof(Resources),"RSC_Kolichestvopovtorovpripere")]
		[LocalizedDescription(typeof(Resources),"RSC_Kolichestvopovtoroveslizna")]
		[LocalizedCategory(typeof(Resources),"RSC_2Obrabotkarblnkov")]
		[Parameter]
		[PropertyOrder(90)]
		public int MaxBetsCount
		{
			get { return _maxBetsCount; }
			set
			{
				_maxBetsCount = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Отменять все ставки после перехода рынка в In-Play
		/// </summary>
		[Parameter]
		[LocalizedCategory(typeof(Resources),"RSC_6Zakrbltierblnkov")]
		[LocalizedDisplayName(typeof(Resources),"RSC_Otmenyatvsestavkiposlepere")]
		[LocalizedDescription(typeof(Resources),"RSC_Esliokazaloschtoprogrammav")]
		[PropertyOrder(5)]
		public bool CancelAllBetsInPlay
		{
			get { return _cancelAllBetsInPlay; }
			set
			{
				_cancelAllBetsInPlay = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Закрывать рынки
		/// </summary>
		[Parameter]
		[LocalizedCategory(typeof(Resources),"RSC_6Zakrbltierblnkov")]
		[LocalizedDisplayName(typeof(Resources),"RSC_Zakrblvatrblnki")]
		[PropertyOrder(10)]
		public bool CloseMarkets
		{
			get { return _closeMarkets; }
			set
			{
				_closeMarkets = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Время закрытия рынков (до старта события)
		/// </summary>
		[Parameter]
		[LocalizedCategory(typeof(Resources),"RSC_6Zakrbltierblnkov")]
		[LocalizedDisplayName(typeof(Resources),"RSC_Vremyazakrbltiyadostartaso")]
		[PropertyOrder(20)]
		public TimeSpan CloseRange
		{
			get { return _closeRange; }
			set
			{
				_closeRange = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Частота попыток закрытия
		/// </summary>
		[Parameter]
		[LocalizedCategory(typeof(Resources),"RSC_6Zakrbltierblnkov")]
		[LocalizedDisplayName(typeof(Resources),"RSC_Chastotapopbltokzakrbltiya")]
		[PropertyOrder(30)]
		public TimeSpan ClosePeriod
		{
			get { return _closePeriod; }
			set
			{
				_closePeriod = value < TimeSpan.FromSeconds(5)
					? TimeSpan.FromSeconds(5)
					: value;

				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Погрешность итогов при уравнивании
		/// </summary>
		[Parameter]
		[LocalizedCategory(typeof(Resources),"RSC_6Zakrbltierblnkov")]
		[LocalizedDisplayName(typeof(Resources),"RSC_Pogreshnostitogovpriuravni")]
		[LocalizedDescription(typeof(Resources),"RSC_Raznitsavitogakhpripopbltk")]
		[PropertyOrder(50)]
		public decimal IgnoreIfWinOnClose
		{
			get { return _ignoreIfWinOnClose; }
			set
			{
				_ignoreIfWinOnClose = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Минимальная сумма закрывающих ставок
		/// </summary>
		[Parameter]
		[LocalizedCategory(typeof(Resources),"RSC_6Zakrbltierblnkov")]
		[LocalizedDisplayName(typeof(Resources),"RSC_Minimalnayasummazakrblvayu")]
		[LocalizedDescription(typeof(Resources),"RSC_Mensheetoysummblzakrblvayu")]
		[PropertyOrder(60)]
		public decimal IgnoreBetAmountOnClose
		{
			get { return _ignoreBetAmountOnClose; }
			set
			{
				_ignoreBetAmountOnClose = value;
				RaiseSettingsChanged();
			}
		}

		/// <summary>
		/// Прибыль/убыток по закрытым рынкам
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Pribbllubbltokpozakrbltblm")]
		[LocalizedCategory(typeof(Resources),"RSC_3Pribbllubbltokpostrategii")]
		[Statistic]
		public decimal ProfitOfClosedMarkets
		{
			get { return _profitOfClosedMarkets; }
			set
			{
				if (_profitOfClosedMarkets == value)
					return;

				_profitOfClosedMarkets = value;
				RaisePropertyChanged("ProfitOfClosedMarkets");
			}
		}

		/// <summary>
		/// Значение тейка или стопа.
		/// </summary>
		[LocalizedDisplayName(typeof(Resources),"RSC_Znacheniestopatikov")]
		[LocalizedDescription(typeof(Resources),"RSC_Znacheniestopavtikakh0stop")]
		[LocalizedCategory(typeof(Resources),"RSC_4Stoploss")]
		[Parameter]
		[PropertyOrder(70)]
		public int StopValue
		{
			get { return _stopValue; }
			set
			{
				_stopValue = value;
				RaiseSettingsChanged();
			}
		}

		[LocalizedDisplayName(typeof(Resources),"RSC_Zashchitnblyspred")]
		[LocalizedDescription(typeof(Resources),"RSC_Zashchitnblyspredvtikakh")]
		[LocalizedCategory(typeof(Resources),"RSC_4Stoploss")]
		[Parameter]
		[PropertyOrder(80)]
		public int StopSafeSpread
		{
			get { return _stopSafeSpread; }
			set
			{
				_stopSafeSpread = value;
				RaiseSettingsChanged();
			}
		}

		[LocalizedDisplayName(typeof(Resources),"RSC_Tipaktivatsii")]
		[LocalizedDescription(typeof(Resources),"RSC_Tipaktivatsiistoplossa")]
		[LocalizedCategory(typeof(Resources),"RSC_4Stoploss")]
		[Parameter]
		[PropertyOrder(90)]
		public StopActivationType StopActivationType
		{
			get { return _stopActivationType; }
			set
			{
				_stopActivationType = value;
				RaiseSettingsChanged();
			}
		}

		[LocalizedDisplayName(typeof(Resources),"RSC_Minimalnayasummavidimosti")]
		[LocalizedCategory(typeof(Resources),"RSC_4Stoploss")]
		[LocalizedDescription(typeof(Resources),"RSC_Minimalnayasummavidimostip")]
		[Parameter]
		[PropertyOrder(100)]
		public decimal StopMinimumAmountOnPrice
		{
			get { return _stopMinimumAmountOnPrice; }
			set
			{
				_stopMinimumAmountOnPrice = value;
				RaiseSettingsChanged();
			}
		}

		#endregion

		#region Звук

		private bool _useAlerts;
		private string _soundFileName;
		private SoundInfo _soundInfo;

		[Parameter]
		[LocalizedCategory(typeof(Resources),"RSC_5Opoveshcheniya")]
		[LocalizedDisplayName(typeof(Resources),"RSC_Vosproizvoditzvukpriprinya")]
		[LocalizedDescription(typeof(Resources),"RSC_Vosproizvoditzvukovoysigna")]
		[PropertyOrder(50)]
		public bool UseAlerts
		{
			get { return _useAlerts; }
			set
			{
				if (_useAlerts == value)
					return;

				_useAlerts = value;
				RaiseSettingsChanged();
				TryCloseSound();
				TryOpenSound();
			}
		}

		[Parameter]
		[LocalizedCategory(typeof(Resources),"RSC_5Opoveshcheniya")]
		[LocalizedDisplayName(typeof(Resources),"RSC_Imyafayla")]
		[LocalizedDescription(typeof(Resources),"RSC_Imyafaylasozvukovblmsignal")]
		[PropertyOrder(60)]
		[Editor(typeof(SoundPathEditor), typeof(SoundPathEditor))]
		public string SoundFileName
		{
			get { return _soundFileName; }
			set
			{
				if (_soundFileName == value)
					return;

				_soundFileName = value;
				RaiseSettingsChanged();
				TryCloseSound();
				TryOpenSound();
			}
		}

		private void TryOpenSound()
		{
			if (File.Exists(_soundFileName) && UseAlerts)
				_soundInfo = SoundPlayer.Open(_soundFileName, false);
		}

		private void TryCloseSound()
		{
			if (_soundInfo == null)
				return;

			SoundPlayer.Close(_soundInfo);
			_soundInfo = null;
		}

		private void TryPlaySound()
		{
			if (_soundInfo != null)
				SoundPlayer.Play(_soundInfo);
		}

		#endregion

		public AutotraderRule()
		{
			BackOffset = new Margin
			{
				Type = MarginTypes.Ticks,
				Value = 1
			};
			LayOffset = new Margin
			{
				Type = MarginTypes.Ticks,
				Value = 1
			};
			IsEnabled = true;
			CloseRange = TimeSpan.FromMinutes(15);
			ClosePeriod = TimeSpan.FromMinutes(1);
			IgnoreIfWinOnClose = 0.05m;
			IgnoreBetAmountOnClose = 0.02m;
			CancelAllBetsInPlay = false;
		}

		/// <summary>
		/// Удалить рынок из обработки.
		/// </summary>
		/// <param name="market">Рынок.</param>
		/// <param name="remove">Добавить рынок на удаление.</param>
		/// <returns>True, если рынок может быть удален, иначе false.</returns>
		public bool TryRemoveMarket(Market market, bool remove)
		{
			if (_marketsToRemove.Contains(market))
				return false;

			if (market.Bets.Any())
				return false;

			if (remove)
				_marketsToRemove.Add(market);

			return true;
		}

		/// <summary>
		/// Обработчик, который вызывается после получения ответа от биржи.
		/// </summary>
		/// <param name="action">Тип действия со ставками.</param>
		/// <param name="bets">Ставки.</param>
		/// <param name="code">Код ошибки.</param>
		/// <returns>True, если необходимо выполнить повторную отправку данных. Иначе, false.</returns>
		protected override RuleBetAction OnAfterSend(BetsAction action, IEnumerable<Bet> bets, ExecutionReportErrorCode? code)
		{
			foreach (var bet in bets)
			{
				if ((action == BetsAction.Place || action == BetsAction.Update) && code == null)
					_bets.SafeGet(bet, k => new BetInfo(k) { NeedStop = true, BetsCount = 1 });

				if (action == BetsAction.Cancel)
					_bets.Remove(bet);
			}

			return RuleBetAction.None;
		}

		/// <summary>
		/// Обработчик, который вызывается при изменении рынка.
		/// </summary>
		/// <param name="market">Рынок.</param>
		protected override void OnMarketChanged(Market market, ApiTypes apiType)
		{
			// рынок перешел в стадию закрытия, обрабатывать не надо
			if (market.NeedToClose)
				return;

			// если рынок добавлен в коллекцию _marketsToRemove, значит его уже обрабатывали ранее и он подлежит удалению
			lock (_sync)
			{
				if (_marketsToRemove.Contains(market))
					return;
			}

			if (AttachBets)
			{
				var newBets = market.Bets.Where(b => !_processedBets.Contains(b) && !_bets.ContainsKey(b)).ToArray();

				foreach (var newBet in newBets)
				{
					if (newBet.State == BetStates.Done || newBet.State == BetStates.Failed)
						_processedBets.Add(newBet);
					else
						_bets.SafeGet(newBet, k => new BetInfo(k) { NeedStop = true, BetsCount = 1 });
				}
			}

			var bets = _bets
				.Values
				.Where(b => b.SourceBet.Market == market)
				.ToArray();

			foreach (var info in bets)
			{
				var bet = info.SourceBet;

				if (AttachBets)
					_processedBets.Add(bet);

				if (bet.UAmount != 0m || bet.MAmount == 0 || bet.State != BetStates.Done)
					continue;

				if (info.NeedStop)
				{
					if (info.Bet == null)
					{
						info.Bet = TryPlaceBet(bet, false, info.BetsCount);
						TryPlaySound();
					}

					if (!TryProcessStop(info))
						continue;

					if (info.Bet != null)
						BetsApi.CancelBet(info.Bet);

					_bets.Remove(bet);
				}
				else
				{
					_bets.Remove(bet);

					TryPlaySound();

					if (info.NeedPlace && (MaxBetsCount == 0 || info.BetsCount < MaxBetsCount))
						TryPlaceBet(bet, true, info.BetsCount + 1);
				}
			}
		}

		private Bet TryPlaceBet(Bet bet, bool needStop, int betsCount)
		{
			Bet newBet = null;

			if (bet.Type == PriceTypes.Lay && Lay)
				newBet = CreateBet(bet, LayOffset.GetPriceWithMargin(bet.AverPrice, -1));

			if (bet.Type == PriceTypes.Back && Back)
				newBet = CreateBet(bet, BackOffset.GetPriceWithMargin(bet.AverPrice, 1));

			if (newBet == null)
				return null;

			_bets.SafeGet(newBet, k => new BetInfo(k) { NeedStop = needStop, BetsCount = betsCount });
			BetsApi.PlaceBet(newBet);

			return newBet;
		}

		private bool TryProcessStop(BetInfo info)
		{
			if (StopValue == 0)
				return false;

			var bet = info.SourceBet;

			switch (StopActivationType)
			{
				case StopActivationType.Koeff:
					{
						var priceItem = bet.Type == PriceTypes.Back
							? bet.Runner.LayPrices.FirstOrDefault()
							: bet.Runner.BackPrices.FirstOrDefault();

						return priceItem != null && TryProcessStop(info, priceItem.Price, priceItem.Amount);
					}
				case StopActivationType.Trade:
					{
						return TryProcessStop(info, bet.Runner.LastPriceMatched, bet.Runner.LastAmountMatched);
					}
			}

			return false;
		}

		private bool TryProcessStop(BetInfo info, decimal price, decimal amount)
		{
			if (amount < StopMinimumAmountOnPrice)
				return false;

			Bet newBet = null;
			var bet = info.SourceBet;

			if (bet.Type == PriceTypes.Back && price >= bet.AverPrice.NextPrice(StopValue))
				newBet = CreateBet(bet, price + StopSafeSpread);

			if (bet.Type == PriceTypes.Lay && price <= bet.AverPrice.PrevPrice(StopValue))
				newBet = CreateBet(bet, price - StopSafeSpread);

			if (newBet == null)
				return false;

			_bets.SafeGet(newBet, k => new BetInfo(k) { NeedStop = false, NeedPlace = false });
			BetsApi.PlaceBet(newBet);

			return true;
		}

		private static Bet CreateBet(Bet bet, decimal price)
		{
			var newBet = new Bet
			{
				Market = bet.Market,
				Runner = bet.Runner,
				Type = bet.Type == PriceTypes.Lay ? PriceTypes.Back : PriceTypes.Lay,
				Price = price,
				UAmount = bet.MAmount
			};

			return newBet;
		}

		/// <summary>
		/// Метод вызывается при сбросе состояния правила.
		/// </summary>
		protected override void OnReseted()
		{
			_marketsToRemove.Clear();
		}

		/// <summary>
		/// Обработчик, который определяет можно ли удалить рынок из обработки.
		/// </summary>
		/// <param name="market">Рынок.</param>
		/// <returns>True, если рынок можно удалить. Иначе False.</returns>
		protected override bool OnCanRemove(Market market)
		{
			lock (_sync)
			{
				if (DateTime.Now - _lastCleanOldMarkets > TimeSpan.FromMinutes(5))
				{
					// не чаще чем каждые 5 минут будем чистить коллекцию _marketsToRemove, удаляя в ней рынки, которые уже точно закончились и больше не попадут в результаты поиска
					var startedMarkets = _marketsToRemove
						.Where(m => m.StartTime.AddMinutes(2) < DateTime.UtcNow)
						.ToArray();

					foreach (var tmpMarket in startedMarkets)
					{
						_processedBets.RemoveWhere(b => b.Market == tmpMarket);
						_bets.RemoveWhere(b => b.Value.SourceBet.Market == tmpMarket);
						_marketsToRemove.Remove(tmpMarket);
					}

					_lastCleanOldMarkets = DateTime.Now;
				}

				if (market.IsInplay && !_marketsToRemove.Contains(market))
				{
					if (CancelAllBetsInPlay)
                        BetsApi.CancelBets(market.Bets.Where(b => b.State == BetStates.Active).ToArray());

					_marketsToRemove.Add(market);
					return true;
				}

				if (market.Status == MarketStatus.Closed)
				{
					_marketsToRemove.Add(market);
					return true;
				}

				if (market.Status == MarketStatus.Suspended && market.StartTime.AddMinutes(2) < DateTime.UtcNow)
				{
					_marketsToRemove.Add(market);
					return true;
				}

				if (_marketsToRemove.Contains(market))
				{
					return true;
				}

				if (!CloseMarkets)
					return false;

				if (market.StartTime - DateTime.UtcNow > CloseRange)
					return false;

				if (!market.NeedToClose)
				{
					market.NeedToClose = true;
					market.LastTryToClose = DateTime.Now;
                    BetsApi.CancelBets(market.Bets.Where(b => b.State == BetStates.Active).ToArray());
					return false;
				}

				if (DateTime.Now - market.LastTryToClose < ClosePeriod)
					return false;

				market.LastTryToClose = DateTime.Now;

				if (!EqualizeMarket(market))
					return false;

				_marketsToRemove.Add(market);
				return true;
			}
		}

		/// <summary>
		/// Функция уравнивает результаты по исходам на рынке
		/// </summary>
		/// <param name="market">Рынок</param>
		/// <returns>True - если результаты по рынку выровнены, False - если нет</returns>
		private bool EqualizeMarket(Market market)
		{
			// проверим результаты по исходам, чтобы понять, закрылся рынок уже или нет
            decimal maxIfWin = -1000000;
            decimal minIfWin = 10000000;
            var runners = market.GetActiveRunners();
            if (runners.Length == 0)
                runners = market.Runners.ToArray();// если рынок уже приостановлен, то и раннеры в состоянии саспендед, поэтому в этом случае возьмем всех раннеров
            if (market.NumberOfWinners > 1)
            {
                foreach (var mRunner in runners)
                {
                    maxIfWin = Math.Max(mRunner.IfWin, maxIfWin);
                    maxIfWin = Math.Max(mRunner.IfLoss, maxIfWin);
                    minIfWin = Math.Min(mRunner.IfWin, minIfWin);
                    minIfWin = Math.Min(mRunner.IfLoss, minIfWin);
                }
            }
            else
            {
                foreach (var mRunner in runners)
                {
                    maxIfWin = Math.Max(mRunner.IfWin, maxIfWin);
                    minIfWin = Math.Min(mRunner.IfWin, minIfWin);
                }
            }

			// если разница между минимальным и максимальным результатами по исходам сильно большая, значит рынок не закрыт
			if ((maxIfWin - minIfWin) <= IgnoreIfWinOnClose)
			{
				// рынок уже закрыт
				// зафиксируем по нему результат в прибыли/убытке работы стратегии
				ProfitOfClosedMarkets += Math.Round(market.Runners.Average(r => r.IfWin), 2);
				return true;
			}

			var bets = market.EqualizeBets().ToList();
			// теперь сравним новые ставки с теми, что уже есть
			// если они одинаковые, то делать не будем, а если отличаются, то отменим уже выставленные
			foreach (var bet in bets.ToArray())
			{
				// удалим закрывающие ставки,
				// которые не попадают под разрешенный диапазон кэфов,
				// это чтобы не делать ставки с к.1000 и 1.01
				if (bet.Price < 1.03m || bet.Price > 950)
				{
					bets.Remove(bet);
					continue;
				}

				// удалим закрывающие ставки,
				// которые не попадают по сумме в разрешенный диапазон
				if (bet.UAmount < IgnoreBetAmountOnClose)
				{
					bets.Remove(bet);
					continue;
				}

				// попробуйем найти непринятые ставки по нужному исходу
				var curBets = from b in market.Bets
							  where b.Runner == bet.Runner
							  && b.UAmount > 0 && b.State == BetStates.Active
							  select b;

				foreach (var curBet in curBets)
				{
					// ставку делать не надо в трех случаях:
					// 1. если кэф новой ставки совпадает с кэфом текущей ставки
					// 2. если кэф новой ставки бэк на один тик меньше кэфа текущей ставки
					// 3. если кэф новой ставки лэй на один тик больше кэфа текущей ставки
					if (bet.Price == curBet.Price)
					{
						// ставка с таким кэфом уже есть, значит новую делать не надо
						bets.Remove(bet);
						break;
					}
					else
					{
						if (bet.Type == PriceTypes.Back
							&& bet.Price == curBet.Price.PrevPrice())
						{
							// наша ставка с лучшим кэфом уже есть, значит новую делать не надо
							bets.Remove(bet);
							break;
						}
						if (bet.Type == PriceTypes.Lay
							&& bet.Price == curBet.Price.NextPrice())
						{
							// наша ставка с лучшим кэфом уже есть, значит новую делать не надо
							bets.Remove(bet);
							break;
						}
						else
						{
							// кэф новой ставки изменился, значит нужно отменить предыдущую
							BetsApi.CancelBet(curBet);
						}
					}
				}
			}

			BetsApi.PlaceBets(bets, false);

			return false;

		}

		/// <summary>
		/// Загрузить настройки.
		/// </summary>
		/// <param name="settings">Настройки.</param>
		public override void Load(Settings settings)
		{
			Lay = settings.GetValue<bool>("Lay");
			Back = settings.GetValue<bool>("Back");
			AttachBets = settings.GetValue<bool>("AttachBets");

			settings.GetValue<Settings>("LayOffset", s => LayOffset.Load(s));
			settings.GetValue<Settings>("BackOffset", s => BackOffset.Load(s));

			CloseMarkets = settings.GetValue<bool>("CloseMarkets", false);
			CloseRange = TimeSpan.Parse(settings.GetValue("CloseRange", "00:15:00"));
			ClosePeriod = TimeSpan.Parse(settings.GetValue("ClosePeriod", "00:01:00"));
			IgnoreIfWinOnClose = settings.GetValue<decimal>("IgnoreIfWinOnClose", 0.05m);
			IgnoreBetAmountOnClose = settings.GetValue<decimal>("IgnoreBetAmountOnClose", 0.02m);
			CancelAllBetsInPlay = settings.GetValue<bool>("CancelAllBetsInPlay", false);

			StopValue = settings.GetValue("StopValue", StopValue);
			StopActivationType = settings.GetValue("StopActivationType", StopActivationType);
			StopSafeSpread = settings.GetValue("StopSafeSpread", StopSafeSpread);
			StopMinimumAmountOnPrice = settings.GetValue("StopMinimumAmountOnPrice", StopMinimumAmountOnPrice);

			UseAlerts = settings.GetValue("UseAlerts", false);
			SoundFileName = settings.GetValue<string>("SoundFileName");

			MaxBetsCount = settings.GetValue("MaxBetsCount", MaxBetsCount);

			base.Load(settings);
		}

		/// <summary>
		/// Сохранить настройки.
		/// </summary>
		/// <param name="settings">Настройки.</param>
		public override void Save(Settings settings)
		{
			settings.SetValue("Lay", Lay);
			settings.SetValue("LayOffset", LayOffset.Save());
			settings.SetValue("Back", Back);
			settings.SetValue("BackOffset", BackOffset.Save());
			settings.SetValue("AttachBets", AttachBets);

			settings.SetValue("CloseMarkets", CloseMarkets);
			settings.SetValue("CloseRange", CloseRange.ToString());
			settings.SetValue("ClosePeriod", ClosePeriod.ToString());
			settings.SetValue("IgnoreIfWinOnClose", IgnoreIfWinOnClose);
			settings.SetValue("IgnoreBetAmountOnClose", IgnoreBetAmountOnClose);
			settings.SetValue("CancelAllBetsInPlay", CancelAllBetsInPlay);

			settings.SetValue("StopValue", StopValue);
			settings.SetValue("StopActivationType", StopActivationType);
			settings.SetValue("StopSafeSpread", StopSafeSpread);
			settings.SetValue("StopMinimumAmountOnPrice", StopMinimumAmountOnPrice);

			settings.SetValue("UseAlerts", UseAlerts);
			settings.SetValue("SoundFileName", SoundFileName);

			settings.SetValue("MaxBetsCount", MaxBetsCount);

			base.Save(settings);
		}

		private void OnPropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			RaiseSettingsChanged();
		}
	}
}
